

#include <oxstd.h>
#include <oxprob.h>
#include <oxfloat.h>
#include <oxdraw.h>

#import <maximize>

static decl
	s_ammData,							// array of data iT x iN x iD (iD = # characteristics)
	s_mParameters,						// matrix of initial parameters
	s_ammParameters,					// pointer to array of parameter matrices  over time
	s_ammTrueParameters,				// pointer to array of true parameter matrices  over time
	s_mTrueStates,						// pointer to array of true parameter matrices  over time
	s_ammPredictedClusterProbabilities,	// matrix iN x iP of posterior mixture probabilities \tau_ij
	s_mInitialPredictedProbabilities,	// time t=1 given t=0 predicted probabilities to start off recursions
	s_ammFilteredClusterProbabilities,	// matrix iT x iN x iP of posterior mixture probabilities \tau_ijt over time
	s_dV,s_dVi,							// degrees of freedom in mv t-densities, and its inverse
	s_iP,								// # of mixture components
	s_dLik,								// the likelihood value
	s_iMeanSpecificA_Parameters = 0,	// if 1, a smoothing parameter per mean, otherwise a single smoothing parameter for all
	s_dRegularizeCovariance = 0,		// this number times a unit matrix is added to the covariance matrices to prevent degenerate results
	s_dFOffset,							// shift of the objective function
	s_iGlobalCount=0, 					// global count to track number of function evaluations
	s_iGlobalCount2=0, 					// global count to track number of function evaluations
	s_iPrint,							// global indicator of amount of information to print
	s_iBurn = 0,						// global integer, number of initial observations to burn in likelihood computations
	s_dWarmStart=1.7;     				// global starting value of the mean smoothing parameter


////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////
//																		  //
// 		             INITIALIZATION AND AUXILIARY ROUTINES                //
//																		  //
////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////

fSwitch(const aA, const aB) {
// switch the contents of aA[0] and aB[0]
//
// INPUTS:
// aA, aB: pointers to objects
	decl aid1;
	aid1 = aA[0]; aA[0] = aB[0]; aB[0] = aid1;
}





fMaxr(const mX)	 {
//	OxMetrics 6 on HPC (High Performance Computing) does not know maxr command!
	decl r, vMaxr=zeros(mX[][0]);
	for (r=0;r<rows(mX);r++) vMaxr[r] = max(mX[r][]);
	return vMaxr;
}

fMinr(const mX)	 {
//	OxMetrics 6 on HPC (High Performance Computing) does not know maxr command!
	decl r, vMinr=zeros(mX[][0]);
	for (r=0;r<rows(mX);r++) vMinr[r] = min(mX[r][]);
	return vMinr;
}




fUseCommandLineArguments(const carglist, ...) {
// set the arguments equal to the command line arguments
	decl args = va_arglist();

	if (sizeof(carglist) == 1) return 0;
	if (sizeof(carglist) != sizeof(args)+1) {
		println("**** ERROR: ARGLIST != CARGLIST ***");
		println("Command line arguments:",carglist);
		println("Function arguments:",args);
		exit(0);
	}
	for (decl i1=1;i1<sizeof(carglist);++i1) if (sscan(carglist[i1],args[i1-1]) != 1) println("Error at ",i1,": ",carglist[i1]," and ",args[i1-1]);
}





fInitializeTimeVaryingParameterStructure(const iT, const iP, const iK, const dScalar=0) {
// Initiale a 3d structure filled with zeros (or an optional scalar).
//
// INPUTS:
// iT	: size of the array
// iP	: rows of the array elements
// iK	: columns of the array elements
// dScalar: [optional] if provided, a scalar, otherwise set to 0
//
// OUTPUTS:
// iT x iP x iK structure filled with dScalar
//
	decl tmpParameters = new array[iT];
	for (decl i1=0;i1<iT;++i1) tmpParameters[i1] = zeros(iP, iK) + dScalar;
	return tmpParameters;
}



fVar3dStructure(const ammX, const vDimensionsToEliminate) {
// Put the covariance matrix of a 3d structure ammX into a matrix
//
// INPUTS:
// ammX: iT x iP x iK array.
//
// OUTPUTS:
// vP: 	1 x (iT * iP * iK) vector, containing vecr(ammParameters[0])'~...~vecr(ammParameters[iT])'
	decl vP = <>;
	for (decl i1=0;i1<sizeof(ammX);++i1) vP ~= vecr(ammX[i1])';
	return vP;
}


	

fVecr3dStructure(const ammX) {
// Put the 3d structure ammX into vecr form, containing a ROW vector
// filled with vecr(ammParameters[0])'~...~vecr(ammParameters[iT])'.
//
// INPUTS:
// ammX: iT x iP x iK array.
//
// OUTPUTS:
// vP: 	1 x (iT * iP * iK) vector, containing vecr(ammParameters[0])'~...~vecr(ammParameters[iT])'
	decl vP = <>;
	for (decl i1=0;i1<sizeof(ammX);++i1) vP ~= vecr(ammX[i1])';
	return vP;
}


	

fReshape3dStructure(const vX, const iT, const iP, const iK) {
// Put the (iT*iP*iK) vector vX into a iT x iP x iK 3d structure mmX,
// where vX = fVecr3dStructure(mmX)
//
// INPUTS:
// vX: 1 x (iT*iP*iK) vector
// iT, iP, iK: desired output dimensions; 
//
// OUTPUTS:
// mmX: iT x iP x iK array.
	if (sizerc(vX) != iT*iP*iK) { println("dimension errors: size is ",sizerc(vX)," versus (",iT,",",iP,",",iK,")"); exit(0); }
	decl mmX = fInitializeTimeVaryingParameterStructure(iT,iP,iK);
	for (decl i1=0;i1<iT*iP*iK;i1+=iP*iK) mmX[i1] = reshape(vX[i1:i1+iP*iK-1],iP,iK);
	return mmX;
}


	

logistic(const vX) {
//Logistic transformation R -> (0,1)
return (vX .>= 0) .? (ones(vX) ./ (ones(vX)+exp(-vX))) .: (exp(vX) ./ (ones(vX)+exp(vX)));	}



logit(const mX) {
//Undoes logistic transformation (0,1) -> R
return log(mX) - log(ones(mX) - mX); }



RestrictPar(const vP) {
//restricts parameters to certain intervals
	decl vPar = vP, idx;
	if (s_iMeanSpecificA_Parameters == 1) idx = range(0,columns(s_ammData[0][][])-1)'; else idx = 0;
//	vPar[idx] = 1.0 * ((vPar[idx] .< 0) .? 0.5 ./ (1 + sqr(vPar[idx])) .: 1 - 0.5 ./ (1 + sqr(vPar[idx])));
//	vPar[idx] = 1*logistic(vPar[idx]);
	return vPar;
}


	
 
////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////
//																		  //
// 		        DRAWING AND PLOTTING ROUTINES                             //
//																		  //
////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////

DrawParametric(const iWindow, const ammParameters)
// Draws parameter paths in P=2 dimensions; only the time varying means.
//
// Inputs:
// iWindow: the window in which the first DrawXMatrix(...) command is performed.
//			(convenient if the plot is part of a multi-panel graph)
// ammParameters: iT x iP x iK array. The evolution over time (iT) is plotted for
//			ammParameters[][ip][0] versus ammParameters[][ip][1] is plotted
//			for ip=0..iP-1, starting from window iWindow
// iNrMeanColumns: the number of columns to plot
//
{
	decl i1,i2,y,x,v,z;	  //i1: time, i2: cluster
	decl iP = rows(ammParameters[0]);         // number of mixture components
	decl iT = sizeof(ammParameters);			// number of time points
	//decl iN = rows(s_ammData[0][][]);	    // number of cross sectional units
	//decl iD = columns(s_ammData[0][][]);	// number of characteristics in data
	
	SetDrawWindow("Time-varying means");
	for (i2=0;i2<iP;++i2)
	{
		y = x = <>;
		for (i1=0;i1<iT;++i1)
		{
			x ~= ammParameters[i1][i2][0];
			y ~= ammParameters[i1][i2][1];
		}
		DrawXMatrix(iWindow,y,0,x,0,2,2+i2);
		DrawAdjust(ADJ_SYMBOL,0,40);
	}
}




DataMovie(...) {
// Play data movie in Window 0
// INPUTS:
// iItem1		:	[optional] vector of integers indicating the characteristics to be plotted; by
//					default plot the first (max) five characteristics (or less if there are less)
// mPosteriorProbabilities : [optional] iN x iP array, iP number of mixture components,
//					containing the posterior probabilities; maximum probability indicates
//					color coding
// NO OUTPUTS
//
// REQUIRED GLOBALS
// s_ammData:      iT x iN x iD array, with iN the number of cross-sectional units,
//               and iT the number of time points, and iD the number of characteristics


	decl args = va_arglist();
	decl iItem1, mExtremes, vColors, vColorScheme = range(1,50);
	decl i1, i2, i3, i4, aid1;


	// set arguments (optional arguments)
	if (sizeof(args) > 0) {
		iItem1 = matrix(args[0]);
	} else {
		iItem1 = range(0,min(columns(s_ammData[0]),5)-1);
	}
	if (sizeof(args) > 1) {
		if (rows(args[1]) == rows(s_ammData[0])) vColors = vColorScheme[maxcindex(args[1]')]';
		else {
			println("\n\n*** ERROR: rows posterior probabilities not equal to ",rows(s_ammData[0])," ***\n");
			vColors = zeros(s_ammData[0][][0]) + vColorScheme[0];
		}
	} else {
		vColors = zeros(s_ammData[0][][0]) + vColorScheme[0];
	}

	// loop over time to determine the dimensions of the plots
	mExtremes = <>;
	for (i1=0;i1<sizeof(iItem1);++i1) {
		for (i2=0;i2<sizeof(s_ammData);++i2) {
			aid1 = deleter(s_ammData[i2][][iItem1[i1]]);
			if (i2>0) {
				mExtremes[i1][] =
					min(mExtremes[i1][0], minc(aid1)) ~
					max(mExtremes[i1][1], maxc(aid1)) ;
			} else {
				mExtremes |= minc(aid1) ~ maxc(aid1);
			}
		}
		mExtremes[i1][] += 0.05 * mExtremes[i1][] * <-1;1> * <-1,1>;
	}


	// loop over time to make the movie
	for (i1=0;i1<sizeof(s_ammData);++i1) {
		for (i2=minc(vColors);i2<=maxc(vColors);++i2) {
			for (i3=0;i3<sizeof(iItem1);++i3) {
				for (i4=0;i4<sizeof(iItem1);++i4) {

					if (i3==i4) continue;
					decl iWindow = i3*sizeof(iItem1) + i4;
			
					DrawXMatrix(iWindow,
						selectifr(s_ammData[i1][][iItem1[i3]],vColors.==i2)',0,
						selectifr(s_ammData[i1][][iItem1[i4]],vColors.==i2)',0,
						1,i2);
					DrawAdjust(ADJ_SYMBOL,2,150);
					DrawAdjust(ADJ_AREA_X,iWindow,mExtremes[i4][0],mExtremes[i4][1],(mExtremes[i4][1]-mExtremes[i4][0])/20);
					DrawAdjust(ADJ_AREA_Y,iWindow,mExtremes[i3][0],mExtremes[i3][1],(mExtremes[i3][1]-mExtremes[i3][0])/20);
				}
			}
		}
		DrawAdjust(ADJ_AREAMATRIX,sizeof(iItem1),sizeof(iItem1));
		ShowDrawWindow();
	}

}


	

////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////
//																		  //
// 		        ESTIMATION AND DIAGNOSTIC ROUTINES                        //
//																		  //
////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////

InfoCriteria(const dlik, const iK, const iPrint)
{
	//numer of obs in the overall panel:
	decl iT = sizeof(s_ammData);			// number of time points
	decl iN = rows(s_ammData[0][][]);	    // number of cross sectional units
	decl iD = columns(s_ammData[0][][]);	// number of characteristics in data
	decl iP = rows(s_ammParameters[0]); 	// number of components
	decl iNum = iT * iN;                    // old command: sumr(sumc(s_mY .== .NaN .? 0 .: 1)); 

	decl AICc =  2*(iK - dlik) + 2*iK*(iK+1)/(iNum - iK - 1);
	decl BIC  = -2*dlik + iK * log(iNum);
	decl mIC  = dlik | iK | AICc | BIC;

	if (iPrint > 1) {
		println("********* AICc and BIC ********");
		println ("\nIC stats ",
	         "%c", {"model"},
			 "%r", {"loglik", "#Par", "AICc", "BIC"},
			 "%cf", {"%12.2f"}, mIC);
	}

	// Calculate Bai and Ng (2002)-type criteria	as well.
	// They might just work in this setting!
	decl i1, i2, dSSEw=0;
	decl mZ = (s_ammFilteredClusterProbabilities .== fMaxr(s_ammFilteredClusterProbabilities)*ones(1,iP) .? 1 .: 0);
	for (i1=0;i1<iT;++i1) for (i2=0;i2<iP;++i2) dSSEw += sumc( sumsqrr(s_ammData[i1][][] - s_ammParameters[i1][i2][:iD-1]).* mZ[][i2] );
	
	//compute Bai and Ng (2002) criteria, in log form. IC2 does best in SW's simulations.
	decl dOF  = dSSEw / (iN*iT*iD);
	decl IC1  = log(dOF) + iP * ((iN+iT)/(iN*iT)) * log((iN*iT)/(iN+iT));
	decl IC2  = log(dOF) + iP * ((iN+iT)/(iN*iT)) * log( fMinr(iN~iT) );
	decl IC3  = log(dOF) + iP * ( log(fMinr(iN~iT))/fMinr(iN~iT) );

	return {AICc, BIC, IC1, IC2, IC3};
		 
}

 

InfoCriteria2(const iPrint)             { 
                //numer of obs in the overall panel: 
                decl i1, i2, i3, i4; 
                decl iT = sizeof(s_ammData);                                      // number of time points 
                decl iN = rows(s_ammData[0][][]);              // number of cross sectional units 
                decl iD = columns(s_ammData[0][][]);    // number of characteristics in data 
                decl iP = rows(s_ammParameters[0]);   // number of components 
  
                decl mZ = (s_ammFilteredClusterProbabilities .== fMaxr(s_ammFilteredClusterProbabilities)*ones(1,iP) .? 1 .: 0); 
  
                //calculate AICk, page 92.  
                decl dSSE=0; 
                for (i1=0;i1<iT;++i1) for (i2=0;i2<iP;++i2) dSSE += sumc( sumsqrr(s_ammData[i1][][] - s_ammParameters[i1][i2][:iD-1]).* mZ[][i2] ); 
                dSSE /= iT; 
                decl dAICk = dSSE + 2*(iN)*iP; 
  
  
                // compute inverse data covariance matrix for Maha distance 
                decl mcov = zeros(iD,iD); 
                for (i1=0; i1<iT; i1++) mcov += variance(s_ammData[i1]);                                                 
                mcov /=               iT; 
                decl mcovi = invertsym(mcov);  //put to unit(iD) for Euclidian distances 
  
                
                //calculate Calinski-Harabsaz Index, as in de Amorim and Hennig (2015), eq (3). 
                decl mDiff, vDiff, vSelect, dSSEW=0; 
                for (i1=0;i1<iT;++i1) 
                for (i2=0;i2<iP;++i2) 
                {              mDiff = s_ammData[i1][][] - s_ammParameters[i1][i2][:iD-1]; 
                                vDiff = sumr((mDiff * mcovi) .* mDiff); 
                                vSelect = vDiff .* mZ[][i2]; 
                                dSSEW += sumc( vSelect ); 
                } 
                dSSEW /= iT; 
  
                decl dSSEb=0; 
                for (i1=0;i1<iT;++i1) 
                for (i2=0;i2<iP;++i2) 
                {              mDiff = s_ammParameters[i1][i2][:iD-1] - meanc(s_ammData[i1][][]);    //distance of estimated means from common data mean, a vector 
                                vDiff = sumr((mDiff * mcovi) .* mDiff); 
                                dSSEb += sumc(mZ[][i2])              * vDiff; 
                } 
                dSSEb /= iT; 
                decl dCHK = (dSSEb / dSSEW) * ((iN - iP)/(iP-1)); 
  
                
                //calculate average Silhouette index, as in de Amorim and Hennig (2015), eq (2). 
                decl dShC=0, vA, vB, vS, vDis, iNum; 
                vS=vA=vB=zeros(iN,1); 
                vDis=zeros(1,iP); 
                for (i1=0;i1<iT;++i1) for (i2=0;i2<iN;++i2) for (i3=0;i3<iP;++i3) if (mZ[i2][i3] == 1)  //this means: firm i2 is located in cluster i3. 
                                {              //average Mahalanobis distance of firm i2 to other firms in the same cluster i3 
                                                mDiff = s_ammData[i1][][] - s_ammData[i1][i2][]; 
                                                vDiff = sumr((mDiff * mcovi) .* mDiff); 
                                                vSelect = vDiff .* mZ[][i3]; 
                                                iNum    = sumc(  mZ[][i3] ) - 1;  //number of firms in cluster i3 other than firm i2 
                                                vA[i2]  += ( 1/iNum ) * sumc( sqrt( vSelect ) ); 
                                                //now, calculate the average distance of firm i2 to firms in other clusters i4 
                                                for (i4=0;i4<iP;++i4) 
                                                {              if (i4 == i3) vDis[i4] = .Inf; 
                                                                if (i4 != i3) 
                                                                {              mDiff = s_ammData[i1][][] - s_ammData[i1][i2][]; 
                                                                                vDiff = sumr((mDiff * mcovi) .* mDiff); 
                                                                                vSelect = vDiff .* mZ[][i4]; 
                                                                                iNum = sumc(mZ[][i4]);   //number of firms in cluster i4 
                                                                                vDis[i4] = (1/ iNum) * sumc( sqrt( vSelect ) ); 
                                                                }; 
                                                } 
                                                //b_i is the shortest average distance to the most neighbouring cluster 
                                                vB[i2] += fMinr(vDis); 
                                } 
                vA /= iT; vB /= iT; 
                vS = (vB-vA) ./ fMaxr(vA~vB); 
                
                
                //calculate Davies-Bouldin Index based on Mahalanobis distance. 
                decl vSi = zeros(1,iP), mM=zeros(iP,iP), mR=zeros(iP,iP); 
                //compute time sum of cluster-specific scatters 
                for (i1=0;i1<iT;++i1) for (i2=0;i2<iP;++i2) 
                {              mDiff = s_ammData[i1][][] - s_ammParameters[i1][i2][:iD-1]; 
                                vDiff = sumr((mDiff * mcovi) .* mDiff); 
                                vSelect = vDiff .* mZ[][i2]; 
                                vSi[i2] +=             (1/sumc(mZ[][i2])) * sumc( sqrt ( vSelect )); 
                }; 
                //compute time sum of distances between clusters 
                for (i1=0;i1<iT;++i1) for (i2=0;i2<iP;++i2) for (i3=0;i3<iP;++i3) 
                {              mDiff = s_ammParameters[i1][i2][:iD-1]               - s_ammParameters[i1][i3][:iD-1]; 
                                vDiff = sumr((mDiff * mcovi) .* mDiff); 
                                mM[i2][i3] +=    sqrt ( vDiff ); 
                }; 
                //compute sum of two scatters / distance ratio for each i and j. Taking the ratio cancels iT. 
                for (i1=0;i1<iP;++i1) for (i2=0;i2<iP;++i2)                mR[i1][i2] = (vSi[i1] + vSi[i2]) / mM[i1][i2]; 
  
                mR = setdiagonal(mR, -.Inf); 
                decl dDBI = meanc( fMaxr(mR) ); 
  
                
                //I return dSSE as well, to compute Hardigan value later in excel, as in de Amorim and Hennig (2015), eq (4). 
                return {-dCHK, dAICk, -meanc(vS), dDBI, dSSE}; 
                                
} 



InfoCriteria3(const iPrint)	 {
	//numer of obs in the overall panel:
	decl i1, i2, i3, i4;
	decl iT = sizeof(s_ammData);			// number of time points
	decl iN = rows(s_ammData[0][][]);	    // number of cross sectional units
	decl iD = columns(s_ammData[0][][]);	// number of characteristics in data
	decl iP = rows(s_ammParameters[0]); 	// number of components

	decl mZ = (s_ammFilteredClusterProbabilities .== fMaxr(s_ammFilteredClusterProbabilities)*ones(1,iP) .? 1 .: 0);

	//calculate AICk, page 92.  
	decl dSSEw=0;
	for (i1=0;i1<iT;++i1) for (i2=0;i2<iP;++i2) dSSEw += sumc( sumsqrr(s_ammData[i1][][] - s_ammParameters[i1][i2][:iD-1]).* mZ[][i2] );
	dSSEw /= iT;
	decl dAICk = dSSEw + 2*(iN*iD)*iP;

	//calculate dCHK, as in de Amorim and Hennig (2015), eq (3).
	decl dSSEb=0;
	for (i1=0;i1<iT;++i1) for (i2=0;i2<iP;++i2)	dSSEb += sumc(mZ[][i2])	* sumsqrr( s_ammParameters[i1][i2][:iD-1] - meanc(s_ammData[i1][][]) );
	dSSEb /= iT;
	decl dCHKu = (dSSEb / dSSEw) * ((iN - iP)/(iP-1));

	//calculate average Silhouette index, as in de Amorim and Hennig (2015), eq (2).
	decl dShC=0, vA, vB, vS, vDis;
	vS=vA=vB=zeros(iN,1);
	vDis=zeros(1,iP);
	for (i1=0;i1<iT;++i1) for (i2=0;i2<iN;++i2) for (i3=0;i3<iP;++i3) if (mZ[i2][i3] == 1)  //this means: firm i2 is located in cluster i3.
		{	//average Euclidian distance of firm i2 to other firms in the same cluster i3
			vA[i2]  += (1/(sumc(  mZ[][i3])-1))*sumc( sqrt( sumsqrr(s_ammData[i1][][] - s_ammData[i1][i2][]) .* mZ[][i3]) );
			//now, calculate the average Euclidian distance of firm i2 to firms in another cluster i4
			for (i4=0;i4<iP;++i4)
			{ 	if (i4 == i3) vDis[i4] = .Inf; 
			  	if (i4 != i3) vDis[i4] = (1/(sumc(mZ[][i4])))*sumc( sqrt( sumsqrr(s_ammData[i1][][] - s_ammData[i1][i2][]) .* (mZ[][i4])) );  
			}
			//b_i is the shortest average distance to the most neighboring cluster
			vB[i2] += fMinr(vDis);
		}
	vA /= iT; vB /= iT;
	vS = (vB-vA) ./ fMaxr(vA~vB);

	//calculate Davies-Bouldin Index based on Euclidian distance; cf. Wikipedia. 
	decl vSi = zeros(1,iP), mM=zeros(iP,iP), mR=zeros(iP,iP);
	//compute time sum of cluster-specific scatters
	for (i1=0;i1<iT;++i1) for (i2=0;i2<iP;++i2)	vSi[i2] +=	(1/sumc(mZ[][i2])) * sumc( sqrt ( sumsqrr(s_ammData[i1][][] - s_ammParameters[i1][i2][:iD-1]).* mZ[][i2] ));
	//compute time sum of Euclidian distances between clusters 
	for (i1=0;i1<iT;++i1) for (i2=0;i2<iP;++i2) for (i3=0;i3<iP;++i3) mM[i2][i3] +=	sqrt ( sumsqrr( s_ammParameters[i1][i2][:iD-1]	- s_ammParameters[i1][i3][:iD-1] ));
	//compute sum of two scatters / distance ratio for each i and j. Taking the ratio cancels iT. 
	for (i1=0;i1<iP;++i1) for (i2=0;i2<iP;++i2)	mR[i1][i2] = (vSi[i1] + vSi[i2]) / mM[i1][i2];

	mR = setdiagonal(mR, -.Inf);
	decl dDBIu = meanc( fMaxr(mR) ); 

	//I return dSSEw as well, to compute Hardigan value later in excel, as in de Amorim and Hennig (2015), eq (4). 
	return {-dCHKu, -meanc(vS), dDBIu};
		 
}






fKMeansClustering(const mData, const iNumberOfClusters, const iNumberOfTries) {
// K-means clustering
//
// Inputs:
// mData    : iN x iD matrix, containing the data
// iNumberOfClusters: integer
// iNumberOfTries: integer, the number of times the algorithm is run from
//            different random initial values / cluster assignments
//
// Outputs:
// mParameters: iNumberOfClusters x iD(ID+3)/2 matrix, rows for each cluster,
//              each row first containing the means, and then the vech of the
//              choleski of the covariance matrix
// mAssignments: iN x iNumberOfClusters matrix, containing a 1 in (i,j) if unit
//               i belongs to cluster j, and zero else

	decl iN = rows(mData);
	decl iD = columns(mData);
	decl mParameters, mAssignments;
	decl dMinimumMSE = .NaN;

	for (decl i0=0; i0<iNumberOfTries; ++i0) {
		// initialize means and variances
		decl mMeans;
		decl amCovs = new array[iNumberOfClusters];

		// random assignment to clusters
		decl mAssignmentsTmp = floor(ranu(iN,1)*iNumberOfClusters);
		decl mAssignmentsPrevious = zeros(mAssignmentsTmp);
		decl iClustersAreChanging = 1;
		decl dMinimumMSETmp;
		
		// iterate until cluster assignment is stable
		while (iClustersAreChanging != 0) {

			// compute the mean
			decl Aid1 = unit(iNumberOfClusters)[mAssignmentsTmp][];
			mMeans = (Aid1'mData) ./ (sumc(Aid1)');

			// compute MSE
			dMinimumMSETmp = sumc(sumr(sqr(mData - mMeans[mAssignmentsTmp][])));

			// update cluster assignments
			Aid1 = zeros(iN,iNumberOfClusters);
			for (decl i1=0;i1<iNumberOfClusters;++i1) Aid1[][i1] = sumr(sqr(mData - mMeans[i1][]));
			mAssignmentsTmp = mincindex(Aid1' )';

			// update convergence tracker
			iClustersAreChanging = sumc(fabs(mAssignmentsTmp - mAssignmentsPrevious));
			mAssignmentsPrevious = mAssignmentsTmp;
		}

		// update to new minimum if the fit is better
		if ((ismissing(dMinimumMSE)) || (dMinimumMSETmp < dMinimumMSE)) {
			dMinimumMSE = dMinimumMSETmp;
			mParameters = mMeans;
			mAssignments = mAssignmentsTmp;
		}
	}

	// compute covariances
	mParameters ~= zeros(iNumberOfClusters, (iD*(iD+1))/2);
	for (decl i1=0;i1<iNumberOfClusters;++i1)
		mParameters[i1][iD:] = vech(choleski(variance(selectifr(mData, mAssignments .== i1))))';

	// return values
	return { mParameters , unit(iNumberOfClusters)[mAssignments][] };
}




fInitializeVp(const iN, const iP, const iD,	const dInverseDegreesOfFreedom) {
	decl mClusterParameters, vP = <>;

	// mean smoothing parameter
	if (s_iMeanSpecificA_Parameters == 1) vP |= ones(iD,1)*s_dWarmStart; else vP |= 1*s_dWarmStart;
	
	// no covariance smoothing parameter
	vP |= <>;

	// starting predicted cluster probabilities as K-Means Clustering assignments
	[mClusterParameters , s_ammPredictedClusterProbabilities[0]] = fKMeansClustering(s_ammData[0], iP, 10);
	s_mParameters = mClusterParameters;
	s_mInitialPredictedProbabilities = s_ammPredictedClusterProbabilities[0];

	// degrees of freedom
	if (dInverseDegreesOfFreedom > 0) vP |= logit( 2 * dInverseDegreesOfFreedom);

	// transition probability decay parameter
	vP |= log( 3 );

	return vP;
}





fGetParameters(vP, const iNrMixtureComponents) {
// Get parameters for the filter from the parameter vector, accounting for several options
//
// INPUTS:
// vP:			parameter vector (dimensions may vary)
// ammData:     iT x iN x iD array, with iN the number of cross-sectional units,
//              and iT the number of time points, and iD the number of characteristics
// iNrMixtureComponents: integer, the number of mixture components
// mPosteriorMixtureProbabilities: (iN x iP) matrix with iN units and iP regimes
// dVi:			inverse degrees of freedom, and degrees of freedom. If dVi==0 or dVi=.NaN, then Gaussian.
//					If dVi<0, the parameter is estimated.
//
// OUTPUTS:
// vEwma: iD+1 x 1 parameter with iD smoothing parameters for the means, and 1 smoothing parameter for variances
// mNewParameters: iP x 0.5*iD*(iD+3) matrix with starting means for each  of the iP regimes, and
//				starting vech(chol(covariances))
//
// REQUIRED GLOBALS
// s_ammData:      iT x iN x iD array, with iN the number of cross-sectional units,
//               and iT the number of time points, and iD the number of characteristics
// s_dVi, s_dV	: inverse degrees of freedom, and degrees of freedom. If s_dVi==0 or s_dVi=.NaN, then Gaussian.
//					If s_dVi<0, the parameter is estimated.


	decl iP,iD, ix=0;
	decl vEwma;

	// transform the parameters
	vP = RestrictPar(vP);

	// get the dimensions
	iD = columns(s_ammData[0]);
	iP = s_iP;
	
	// first element(s) vP: smoothing parameter(s) for the mean
	if (s_iMeanSpecificA_Parameters == 1) {
		vEwma = vP[ix:ix+iD-1]; ix+=iD;
	} else {
		vEwma = ones(iD,1) .* vP[ix:ix]; ix+=1;
	}
	
	// no element for: smoothing parameter(s) for the covariance matrix
	vEwma |= 0;

	// next element vP: degrees of freedom
	if (ismissing(s_dVi) || (s_dVi == 0)) {
		s_dV = .NaN;
	} else if (s_dVi > 0) {
		s_dVi = 0.5 * logistic(vP[ix]); ix+=1;
		s_dV = 1.0 / s_dVi;
	} else {
		s_dV = 1.0 / fabs(s_dVi);
	}

	decl dGamma = exp(vP[ix]); ix+=1;
	
	return {vEwma, dGamma};
}



fComputeDistanceMatrix(const mClusterMeans, const mClusterChols, const dGamma) {
	decl iP = rows(mClusterMeans);
	decl mDistanceMatrix = zeros(iP,iP);
	decl mChol  = lower(unvech(meanc(mClusterChols)));
	decl mCholi = invert(mChol)';

	for (decl i1=0;i1<iP;++i1) {
		for (decl i2=i1+1;i2<iP;++i2) {
			// compute the distance between means, then make symmetric
			mDistanceMatrix[i1][i2] = sqrt(sumsqrr( (mClusterMeans[i1][] - mClusterMeans[i2][]) * mCholi ));
			mDistanceMatrix[i2][i1] = mDistanceMatrix[i1][i2];
		}
	}

	return mDistanceMatrix;
}



fMapClusterMeanDistancesIntoProbabilities(const mDistanceMatrix, const dGamma) {
// Map the cluster means distances into a transition probability matrix.
// The underlying model is
//     Pi(l,m) = exp(-dGamma*d(l,m)) / (1 + sum_(q\not= l) exp(-dGamma*d(l,q)))
// with d(l,q) is the distance between means of cluster l and q.
//
// INPUTS:
// mClusterMeans : iP x iD matrix with the iP cluster means in dimension iD
// dGamma      : scalar, the rate of decay of the transition probabilities from one cluster to the other
//
// OUTPUTS:
// mTransitionMatrix : iP x iP matrix containing transition probabilities

	decl iP = rows(mDistanceMatrix);
	decl mTransitionMatrix = mDistanceMatrix;

	// transform distances in cluster means into transition probabilities
	mTransitionMatrix = exp(-dGamma * mTransitionMatrix);
	mTransitionMatrix ./= sumr(mTransitionMatrix);
	return mTransitionMatrix;
}





fGetMeanAndCovarianceMatrixFromParameter(const mParameters, const iD, const s_dRegularizeCovariance=0) {
	decl ix, vMean, mCov, mChol;

	ix = 0;
	vMean = mParameters[ix:ix+iD-1];	//vMean refers to component i2 at time i1, of dimension iD x iD
	ix += iD;
	mChol = mParameters[ix:ix+iD*(iD+1)/2-1];
	ix += iD*(iD+1)/2;
	mChol = lower(unvech(mChol));
	mCov = mChol * mChol' + s_dRegularizeCovariance * unit(iD);
	mChol = choleski(mCov);
	return {vMean, mCov, mChol};
}




fHMMGasFilter(mParameters, const vPar, const dGamma, const mInitialPredictedClusterProbabilities, const dSmooth) {
//
// Run the GAS filter for the mean and the covariance matrix.
// At the moment, the filter is a single parameter EWMA filter.
//
// ORDER OF THINGS
// Compute the likelihood for the HMM dynamic mixture model. The logistics of the
// computations are as follows:
// 1...starting from initial predicted mixture probabilities, means, and variances for each
// mixture component, and setting t=1
// 2...compute the likelihood contribution for t and add to likelihood
// 3...compute the transition probabilities from t to t+1 using the means for time t 
// 4...compute the score of the time t density to obtain scores using the means and
// variances, and use the scores to update the means and covariances to those for t+1
// 5...use the transition probabilities from step 3 to update the predicted probabilities
// to time t+1
// 6...increase t to t+1 and go to step 2 untill all times are processed
//
// INPUTS:
// mParameters : iP x (iD + 0.5*iD*(iD+1)) array, with iP the number of mixture components.
//               The columns mParameters[iP][] are
//                [0:iD-1]: initial means
//                [iD:iD+0.5*iD*(iD+1)-1]: vech(choleski of initial cov matrix)'
// vPar	       : (iD+1)x1 double, the GAS smoothing parameter for mean (0..iD-1) and variance (iD element).
// dGamma      : scalar, the rate of decay of the transition probabilities from one cluster to the other
//               in terms of the distance between cluster means, using
//               Pi(l,m) = exp(-dGamma*d(l,m)) / (1 + sum_(q\not= l) exp(-dGamma*d(l,q)))
//               and d(l,q) is the distance between means of cluster l and q.
//
// REQUIRED GLOBALS
// s_ammData:      iT x iN x iD array, with iN the number of cross-sectional units,
//               and iT the number of time points, and iD the number of characteristics
// s_mParameters: iP x (iD + 0.5*iD*(iD+1)) array, with iP the number of mixture components.
//               The columns mParameters[iP][] are
//                [0:iD-1]: initial means
//                [iD:iD+0.5*iD*(iD+1)-1]: vech(choleski of initial cov matrix)'
// s_ammTransitionProbabilities: iT x iP x iP array structure on output containing the transition probabilities 
//				 for time iT from cluster *row* to cluster *column*
// s_dVi, s_dV	: inverse degrees of freedom, and degrees of freedom. If s_dVi==0 or s_dVi=.NaN, then Gaussian.
//					If s_dVi<0, the parameter is estimated. These are the values from the previous iteration,
//					needed for example when running the filter itself in the E-step. For the computation of likelihoods and
//					the objective function, however, we need the degrees of freedom of the current iteration.
//					This is particularly relevant for the M-step.
//
// OUTPUTS:
// vLik        :  iT vector holding the log densities (i.e., likelihood
//                contributions) for each time point

	decl iD,iN,iT,iP,i1,i2,ix,aid1;
	decl vMean, mChol, mCov, mCovi, vLik, mLikTimeT, mPi,
		mFilteredClusterProbabilities, mPredictedClusterProbabilities, mClusterDistances,
		Score;
//	decl dSmooth = 0.25;

	// initialize
	iT = sizeof(s_ammData);			// number of time points
	iN = rows(s_ammData[0][][]);	// number of cross sectional units
	iD = columns(s_ammData[0][][]);	// number of characteristics in data
	iP = rows(mParameters);         // number of mixture components
	mLikTimeT = zeros(iN,iP);		// matrix of likelihood contributions at time t firm i cluster j
	vLik = zeros(1,iT);				// vector of likelihood contributions for each time t
	mPredictedClusterProbabilities = mInitialPredictedClusterProbabilities;

	// run the filter per time step
	for (i1=0;i1<iT;++i1) {

		// store the time varying mean and time varying scale parameters
		s_ammParameters[i1] = mParameters;
		s_ammPredictedClusterProbabilities[i1] = mPredictedClusterProbabilities;

		// compute and store time t (i1) likelihood contributions
		vLik[i1] = 0;
		for (i2=0;i2<iP;++i2) {
		
			// get out the parameters
			[vMean, mCov, mChol] = fGetMeanAndCovarianceMatrixFromParameter(mParameters[i2][], iD, s_dRegularizeCovariance);
			mCovi = invertsym(mCov);

			// using the filtered cluster probabilities, compute the log of the cluster j likelihood contribution
			if (ismissing(s_dVi) || (s_dVi == 0)) {
				// Gaussian case
				mLikTimeT[][i2] = - 0.5*iD*log(M_2PI)											
							 - sumr(log(diagonal(mChol)))
							 - sumr(((s_ammData[i1][][] - vMean) * mCovi) .* (s_ammData[i1][][] - vMean));
			} else {
				// Student's t case
				mLikTimeT[][i2] =  loggamma( (s_dV + iD)/2 )
							 - loggamma( s_dV/2 )
							 - 0.5*iD*log(s_dV * M_PI)											
							 - sumr(log(diagonal(mChol)))
							 - 0.5 * (s_dV + iD) * log(1 + fabs(s_dVi) * sumr(((s_ammData[i1][][] - vMean) * mCovi) .* (s_ammData[i1][][] - vMean)));
			}
		}


		// Compute log likelihood contributions across states for time t.
		// This is the log of the weighted sum of the exp(log likelihoods for time t and cluster j).
		// The addition is numerically sensitive, and  therefore we use a standard
		// numerical scaling 'trick'.
		aid1 = maxr( mLikTimeT );
		mFilteredClusterProbabilities =  // intermediate storage object; to be updated later
				// for numerical reasons, subtract the largest component log likelihood for
				// each individual from the log likelihood values before exponentiation
				exp(mLikTimeT - aid1) .* mPredictedClusterProbabilities;
		vLik[i1] =
			// sum across individuals
			sumc(
				// for numerical reasons, subtract the largest component likelihood for
				// each individual from the log likelihood values before exponentiation
				aid1
				+
				// log probability of full (predicted) mixture density
				// using the intermediate storage object
				log(sumr( mFilteredClusterProbabilities ))
			);
			
		// compute the filtered probabilities, to be used in the GAS update later on
		mFilteredClusterProbabilities ./= sumr(mFilteredClusterProbabilities);
		s_ammFilteredClusterProbabilities[i1] = mFilteredClusterProbabilities;
		
		if (i1==0) mClusterDistances = fComputeDistanceMatrix(mParameters[][:iD-1], mParameters[][iD:], dGamma); 
		else mClusterDistances =  (1-dSmooth)*mClusterDistances + dSmooth*fComputeDistanceMatrix(mParameters[][:iD-1], mParameters[][iD:], dGamma);
				
		// compute the transition probabilities from t to t+1 based on the
		// means from time t
		mPi = fMapClusterMeanDistancesIntoProbabilities(mClusterDistances, dGamma);
			

		// loop over clusters to update the cluster means and variances
		for (i2=0;i2<iP;++i2) {

			// get out the parameters
			[vMean, mCov, mChol] = fGetMeanAndCovarianceMatrixFromParameter(mParameters[i2][], iD, s_dRegularizeCovariance);
			mCovi = invertsym(mCov);
			
			// compute the mean's score step and update the mean
			Score = mFilteredClusterProbabilities[][i2]'(s_ammData[i1] - vMean) /
					sumc( mFilteredClusterProbabilities[][i2] );
			
			vMean += vPar[:iD-1]' .* Score;
			// compute the covariance matrix's score step and update the covariance matrix
			Score = (s_ammData[i1] - vMean)' *
					(mFilteredClusterProbabilities[][i2] .* (s_ammData[i1] - vMean)) /
					sumc(mFilteredClusterProbabilities[][i2]) - mCov;
			mCov  += vPar[iD] * Score;

			// store parameters into matrix object
			ix = 0;
			mParameters[i2][ix:ix+iD-1] = vMean;
			ix += iD;
			mParameters[i2][ix:ix+iD*(iD+1)/2-1] = vech(choleski(mCov))';
			ix += iD*(iD+1)/2;
		}

		// update predicted probabilities for next period
		mPredictedClusterProbabilities = mFilteredClusterProbabilities * mPi;

	}

	// update initial state probabilities
//	if (minr(meanc(s_ammPredictedClusterProbabilities[0])) > 0.1) 
//		s_mInitialPredictedProbabilities = s_ammFilteredClusterProbabilities[0];

	// plot the time varying parameter path
//	DrawParametric(0, s_ammParameters); DrawTitle(0,sprint("\\alpha=",vPar[0],";  \\gamma=",dGamma,"\n",mPi)); ShowDrawWindow();

	return vLik;

}



fMaximumLikelihood(const iNrMixtures, const iNrStartingValues, const iPrint, const dInverseDegreesOfFreedom, const dSmooth) {
//
// REQUIRED GLOBALS
// s_ammData:      iT x iN x iD array, with iN the number of cross-sectional units,
//               and iT the number of time points, and iD the number of characteristics
// s_mParameters:iP x (iD + 0.5*iD*(iD+1)) array, with iP the number of mixture components.
//               The columns mParameters[iP][] are
//                [0:iD-1]: initial means
//                [iD:iD+0.5*iD*(iD+1)-1]: vech(choleski of initial cov matrix)'
// s_dVi, s_dV	: inverse degrees of freedom, and degrees of freedom. If s_dVi==0 or s_dVi=.NaN, then Gaussian.
//					If s_dVi<0, the parameter is estimated.

	decl iP = iNrMixtures;
	decl iD = columns(s_ammData[0]);
	decl iN = rows(s_ammData[0]);
	decl iT = sizeof(s_ammData);
	decl vP, dF=0, vEwma, mResults = <>;
	decl dAIC, dBIC;

	// initialize
	s_iP = iP;
	// shift in objective function (numerical issue)
	s_dFOffset = 0;
	
	//Iteratation steps printing:
	if (iPrint > 4) MaxControl(250, 1, 0); else MaxControl(250, 0, 1);
	MaxControlEps(1e-4, 1e-3);
	
	// loop over number of staring values
	for (decl iS=0;iS<iNrStartingValues;iS++) {

		// initialize
		s_dVi = dInverseDegreesOfFreedom;
		vP = fInitializeVp(iN, iP, iD, dInverseDegreesOfFreedom);
		s_mInitialPredictedProbabilities = s_ammPredictedClusterProbabilities[0];

		//
		// define the local likelihood function as a lambda function
		//
		decl fLocalLikelihoodFunction =
			[=](const vP, const adF, const avG, const amH){

				decl vEwma, dGamma, vLik;
				if (maxc(fabs(vP)) > 50) return 0;

				// transform the parameter vector into the necessary parameters
				[vEwma, dGamma] = fGetParameters(vP, iP);
				if (fabs(log(dGamma)) > 5) return 0;

				// run the GAS - HMM - cluster filter
				vLik = fHMMGasFilter(s_mParameters, vEwma, dGamma, s_mInitialPredictedProbabilities, dSmooth);
				adF[0] = meanr(vLik[s_iBurn:]);

				// check whether likelihood can be evaluated
				if (ismissing(adF[0])) return 0; else return 1;
			};

		// run the maximization of the likelihood
		MaxControl(20,-1);
		decl ir = MaxBFGS(fLocalLikelihoodFunction, &vP, &dF, 0, 1);
		// set the initial predicted probabilities to the t=1 filtered probabilities
		MaxControl(100,-1);
		s_mInitialPredictedProbabilities = s_ammFilteredClusterProbabilities[0];
		ir = MaxBFGS(fLocalLikelihoodFunction, &vP, &dF, 0, 1);
		println("BFGS exit status: ", ir);

		// make intermediate plot
		if (iPrint > 2) {DrawParametric(0, s_ammParameters); ShowDrawWindow();}
	
		//report info criteria
		decl vModelSelectionCriteria = <>, aid1, aid2, aid3, aid4, aid5;

		// switch regimes around if observation 1 does not have the highest tau
		// note that the time paths have not been changed; for this the filter needs to be rerun!
		if (iP==2) if (s_ammFilteredClusterProbabilities[0][0][0] < s_ammFilteredClusterProbabilities[0][0][1]) {
			// change the regime probabilities
			for (decl i4=0;i4<iT;++i4) {
				s_ammFilteredClusterProbabilities[i4] = s_ammFilteredClusterProbabilities[i4] * <0,1;1,0>;
				s_ammParameters[i4] = <0,1;1,0> * s_ammParameters[i4];
			}
		}
	
		// store results for this starting value
 
		mResults |=	dF ~ ir ~ vP' ~ RestrictPar(vP)' ~
					fVecr3dStructure(s_ammParameters) ~
					fVecr3dStructure(s_ammFilteredClusterProbabilities) ~
					vModelSelectionCriteria;
	}																						   

	// choose result for best starting value and evaluate objective function once more to
	// set time varying parameter path
	decl vResults = selectifr(mResults, mResults[][0] .== maxc(mResults[][0]) .? 1 .: 0)[0][];
	return vResults ~ fVecr3dStructure(s_ammTrueParameters) ~ vecr(s_mTrueStates)' ~ fVecr3dStructure(s_ammData);
	
}




////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////
//																		  //
// 		        SIMULATION PLOTTING ROUTINES                              //
//																		  //
////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////

fSimulateData3(const iNrFirmsPerMixtureComponent, const iT, const iNrMixtures, const dInverseDegreesOfFreedom, 
		const dCircleRadius, const dCircleMidpointDistance, const iNrLaps, const iTVprob, const dGamma=50, const iPlot=1,
		const dSemiMajorAxis=1, const iEllipse1Orientation=1, const iEllipse2Orientation=1,
		const dVarianceComponent1=1, const dVarianceComponent2=1, const dCovariance=0, const dSmooth=0.1, const iSimulationType=0) {

// simulate data in concentric circles with increased midpoint distance 
//
// INPUTs:
// iPlot: integer; if 1, then plot the data while generating
// iSimulationType: integer; 0 for ovals, 1 for converging circles, 2 for Y-shaped and 3 for flipped-Y. It can make some of the other inputs redundant.
//
// NO OUTPUTS
// global s_ammData

	decl iN = iNrFirmsPerMixtureComponent, iP = iNrMixtures;		// iN = firms per cluster!
	decl iD = 2;					 								// magic number: #firm characteristics
	decl iTimeAhead = ceil(iT/4);		 							// factor by how much the second cluster is ahead of the first
	decl vTime = iNrLaps*M_2PI * range(1,iT + iTimeAhead) / iT, vMean;
	decl mX, mY, mZ, mAid, mChol, mChol2, iOffset;
	decl mCorr = ( dVarianceComponent1 ~ -dCovariance ) | (-dCovariance ~ dVarianceComponent2 );		// magic numbers; covariance matrix of clusters
	decl mStateTransitionProbabilities = zeros(iP, iP), mPreviousMeans, vCurrentStates, mCircleCenters, mClusterDistances;
	decl dCurrentX, dCurrentY; // Used to make some computations clearer.

	// initialize
	mX = mY = zeros(iP*iN,iT);
	vCurrentStates = range(0,iP-1)' ** ones(iN, 1);	 // a iN*iP column vector with the indices of the clusters from 0 to iP-1
	s_ammData = new array[iT];
	mChol = choleski(mCorr)';
	mChol2 = choleski(mCorr .* (2*unit(iD)-ones(iD, iD)))'; // Flip the sign of the off-diagonals

	// initialize (previous) means
	mCircleCenters = ( 0 ~ 0 ) | (dCircleMidpointDistance ~ 0);	// One row for each cluster, one col for each dimension

	// Depending on what type of simulation is desired, generate means
	if (iSimulationType==1){ 
		// Converging circles
		vTime = M_2PI * range(1,iT) / (2*iT); // Fix angles to half a rotation.
		// Overwrite the initial means using two circles side by side
		mCircleCenters = ( -dCircleRadius ~ 0 ) | (dCircleRadius ~ 0);
		mPreviousMeans = mCircleCenters + dCircleRadius * (
				((0) ~ (1)) |
				((0) ~ (1)));
	} else 	if (iSimulationType==2){
		// Y-shaped
		// Here the radius becomes the Y distance and the dCircleMidpointDistance becomes the X distance.
		// No more circular movement, just straight lines.
		mCircleCenters = ( dCircleMidpointDistance ~ dCircleRadius ) | (-dCircleMidpointDistance ~ dCircleRadius);
		mPreviousMeans = mCircleCenters ;
	} else 	if (iSimulationType==3){
		// flipped-Y 
		// Same as 2 but starts together and separates
		mCircleCenters = ( 0 ~ dCircleRadius ) | (0 ~ dCircleRadius); // Start in the same position, will split after t=T/2
		mPreviousMeans = mCircleCenters ;
	} else 	if (iTVprob==1 ) {
			// Elliptical trajectories 
			mPreviousMeans = mCircleCenters + dCircleRadius * (
				(((-(dSemiMajorAxis^(1 - iEllipse1Orientation)) * cos(0)) ~
					(dSemiMajorAxis^(iEllipse1Orientation)) * sin(0)) |
				((dSemiMajorAxis^(1 - iEllipse2Orientation)) * cos(vTime[iTimeAhead]) ~
					((dSemiMajorAxis^(iEllipse2Orientation)) * sin(vTime[iTimeAhead])))));
	} else
			// Synchronized elliptical trajectories 
			mPreviousMeans = mCircleCenters + dCircleRadius * (
				(((-(dSemiMajorAxis^(1 - iEllipse1Orientation)) * cos(0)) ~
					(dSemiMajorAxis^(iEllipse1Orientation)) * sin(0)) |
				(-(dSemiMajorAxis^(1 - iEllipse2Orientation)) * cos(vTime[iTimeAhead]) ~
					((dSemiMajorAxis^(iEllipse2Orientation)) * sin(vTime[iTimeAhead])))));
	
	for (decl i1=0;i1<iT;++i1) {
		// To change the starting position of cluster 2, pretend that it's in the future
		decl i1c2 = i1 + iTimeAhead;

		// store true parameters
		s_ammTrueParameters[i1] = mPreviousMeans ~ (vech(mChol)' | vech(mChol2)'); // vech uses the lower diagonal so the difference between mChol and mChol2 is not captured.
		s_mTrueStates[i1][] =  vCurrentStates';
		// first, generate the firm observations
		mAid  = (vCurrentStates .* rann(iN*iP,iD) * mChol) + ((1-vCurrentStates) .* rann(iN*iP,iD) * mChol2);
		// check whether we generate from a normal or Student's t
		if (ismissing(dInverseDegreesOfFreedom) || (dInverseDegreesOfFreedom == 0)) {} else {
			// generate Student t
			mAid ./= sqrt( fabs(dInverseDegreesOfFreedom) * ranchi(iN,1,1/fabs(dInverseDegreesOfFreedom)) );
		}
		// store the data for this cluster
		s_ammData[i1] = mPreviousMeans[vCurrentStates][] + mAid;
		// if plot required, make it
		if (iPlot>=1) {
			for (decl i2=0;i2<iP;++i2) {
				// draw data for this cluster using specific color / symbol
				DrawXMatrix(0,s_ammData[i1][i2*iN:i2*iN+iN-1][1]', 0,
							  s_ammData[i1][i2*iN:i2*iN+iN-1][0]', 0,
							  1, 2+i2);
				DrawAdjust(ADJ_SYMBOL,2,150);
				decl dmin = (dCircleRadius + iP * dCircleMidpointDistance) + 3 * sqrt(max(diagonal(mCorr)));
				DrawAxis(0,1,0,-dmin,dmin,-dmin,dCircleMidpointDistance,1,0);
				DrawAxis(0,0,0,-dmin,dmin,-dmin,dCircleMidpointDistance,1,0);
				// Also draw means in green (6)
				DrawXMatrix(0,mPreviousMeans[i2][1]', 1,
							  mPreviousMeans[i2][0]', 1,
							  1, 6);
			}
		}
		

		// second, generate transition probability matrix
 		if (i1==0) mClusterDistances = fComputeDistanceMatrix(mPreviousMeans, ones(iP,1) * vech(unit(iP))', dGamma);  
		else mClusterDistances =  (1-dSmooth)*mClusterDistances + dSmooth*fComputeDistanceMatrix(mPreviousMeans, ones(iP,1) * vech(unit(iP))', dGamma);		
		// compute the transition probabilities from t to t+1 based on the
		// means from time t
		mStateTransitionProbabilities = fMapClusterMeanDistancesIntoProbabilities(mClusterDistances, dGamma);


		
		// third, generate firm new states
		mAid =  sumr( ranu(iN * iP, 1) .> (cumulate(mStateTransitionProbabilities' )')[vCurrentStates][] );
		if (iPlot>=1) {
			DrawTitle(0,sprint("State changes: ",sumc(vCurrentStates .!= mAid),"\n",mStateTransitionProbabilities, "\n", "Time: ", i1, " / ", iT));
 			ShowDrawWindow();
		}
		vCurrentStates = mAid;

		
		// fourth, update the means
		if (iSimulationType==1){
			if (i1 < iT/2) {
				mPreviousMeans = mCircleCenters + dCircleRadius * (
					((sin(vTime[i1])) ~ (cos(vTime[i1])) |
					((-sin(vTime[i1])) ~ (cos(vTime[i1])))));
			} // Else nothing, just stay put
		} else 	if (iSimulationType==2){
			// Take linear steps
			dCurrentX = ((i1+1 < iT/2) * 2*(i1+1)*dCircleMidpointDistance/iT) + ((i1+1 >= iT/2) * dCircleMidpointDistance);
			mPreviousMeans = mCircleCenters	+  (
				(( -dCurrentX ) ~ (-(i1+1)*2*dCircleRadius/iT)) |
				((  dCurrentX ) ~ (-(i1+1)*2*dCircleRadius/iT)));
		} else 	if (iSimulationType==3){
			// Take linear steps
			dCurrentX = ((i1+1 >= iT/2) * 2 * (i1+1 - (iT/2))*dCircleMidpointDistance/iT); //+ ((i1+1 < iT/2) * dCircleMidpointDistance);
			mPreviousMeans = mCircleCenters	+  (
				(( -dCurrentX ) ~ (-(i1+1)*2*dCircleRadius/iT)) |
				((  dCurrentX ) ~ (-(i1+1)*2*dCircleRadius/iT)));
				
		} else 	if (iTVprob==1 ) {
			mPreviousMeans = mCircleCenters + dCircleRadius * (
				(((-(dSemiMajorAxis^(1 - iEllipse1Orientation)) * cos(vTime[i1])) ~
					(dSemiMajorAxis^(iEllipse1Orientation)) * sin(vTime[i1])) |
				((dSemiMajorAxis^(1 - iEllipse2Orientation)) * cos(vTime[i1c2]) ~
					((dSemiMajorAxis^(iEllipse2Orientation)) * sin(vTime[i1c2])))));
		} else
			mPreviousMeans = mCircleCenters + dCircleRadius * (
				(((-(dSemiMajorAxis^(1 - iEllipse1Orientation)) * cos(vTime[i1])) ~
					(dSemiMajorAxis^(iEllipse1Orientation)) * sin(vTime[i1])) |
				(-(dSemiMajorAxis^(1 - iEllipse2Orientation)) * cos(vTime[i1c2]) ~
					((dSemiMajorAxis^(iEllipse2Orientation)) * sin(vTime[i1c2])))));
	}
	 	return s_ammData;
}






////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////
//																		  //
// 		                      MAIN                                        //
//																		  //
////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////

main() {

	// initializations
	decl aProgramArguments = arglist();
	decl iNrSimulations =250, iSim, sName;
	decl vResult, mResult = <>;
	decl dInverseDegreesOfFreedomForSimulation = 0, 		dInverseDegreesOfFreedomForEstimation = 0;
	decl iNrMixtureComponentsForSimulation=2,				iNrMixtureComponentsForEstimation=3;
	decl dCircleRadius=2, dCircleMidpointDistance=2, iNrLaps=1;
	decl iN=100, iT=40, iD=2; // iN is number of firms PER MIXTURE COMPONENT; total nr. firms is iN*iNrMixtureComponentsForSimulation
	decl iNrStartingValuesForEM = 1, iRanSeed = <>;
	decl dGamma=0.7;
	decl iTVprob=1;	   // if 0, transition probs are constant; if 1, they are time-varying. This alters the trajectories of the true means.
	decl dSimSemiMajorAxis = 2, dSimVarianceComponent1 = 1, dSimVarianceComponent2 = 1,	dCovariance=0.9;
	decl sResultsPath="results_v15";
	decl iEllipse1Orientation=1; // 1 for standing; 0 for lying down
	decl iEllipse2Orientation=0; // 1 for standing; 0 for lying down
	decl dSmoothEstimation=0.25;
	decl dSmoothDGP=0.1;
	decl iSimulationType=3; // 0 for ovals, 1 for converging circles, 2 for Y-shaped and 3 for flipped-Y.
	
	s_iPrint = 1;
	


	// overwrite global choices with program arguments list
	fUseCommandLineArguments(aProgramArguments
		,&iRanSeed
		,&iSimulationType
		,&iN
		,&iT
		,&dGamma
		,&dInverseDegreesOfFreedomForSimulation
		,&dInverseDegreesOfFreedomForEstimation
		,&dCircleRadius
		,&dCircleMidpointDistance
		,&iNrLaps
		,&iNrMixtureComponentsForEstimation
		,&iTVprob
		,&s_iMeanSpecificA_Parameters
		,&dSimSemiMajorAxis
		,&iEllipse1Orientation
		,&iEllipse2Orientation
		,&dSimVarianceComponent1
		,&dSimVarianceComponent2
		,&dCovariance
		,&dSmoothEstimation
		,&dSmoothDGP
	);
	println("\n\n=================================\nSeed:",iRanSeed,"\n(iN,iT,iD) = (",iN,",",iT,",",iD,")\n(iPsim,iPest) = (",iNrMixtureComponentsForSimulation,",",iNrMixtureComponentsForEstimation,")\n(DFiSim, DFiEst) = (",dInverseDegreesOfFreedomForSimulation,",",dInverseDegreesOfFreedomForEstimation,")\n(Radius,Dist,Laps) = (",dCircleRadius,",",dCircleMidpointDistance,",",iNrLaps,")\n(#Sims,#Starts) = (",iNrSimulations,",",iNrStartingValuesForEM,")\n s_iMeanSpecificA_Parameters = ", s_iMeanSpecificA_Parameters,"\n\n============================\n");


	// initialize necessary globals
	s_ammParameters = fInitializeTimeVaryingParameterStructure(iT,iNrMixtureComponentsForEstimation,0.5*iD*(iD+3));
	s_ammTrueParameters = fInitializeTimeVaryingParameterStructure(iT,iNrMixtureComponentsForSimulation,0.5*iD*(iD+3));
	s_mTrueStates = zeros(iT,iNrMixtureComponentsForSimulation*iN);


	// initialize
	s_ammPredictedClusterProbabilities = fInitializeTimeVaryingParameterStructure(iT,iNrMixtureComponentsForEstimation*iN,iD,1.0/iNrMixtureComponentsForEstimation);
	s_ammFilteredClusterProbabilities = fInitializeTimeVaryingParameterStructure(iT,iNrMixtureComponentsForEstimation*iN,iD,1.0/iNrMixtureComponentsForEstimation);
	
	// run simulations
	if (rows(iRanSeed) > 0) {println("Setting the random seed to ",iRanSeed); ranseed(iRanSeed);}
	for (iSim=0;iSim<iNrSimulations;++iSim) {

		s_iGlobalCount2 = iSim;

		// sign of life
		if (fmod(iSim,1)==0) {
			println("\n==================================");
			println("Simulation: ",iSim);
			println("==================================");
		}					   
		
		// simulation
		fSimulateData3(iN, iT, iNrMixtureComponentsForSimulation, dInverseDegreesOfFreedomForSimulation,
						dCircleRadius, dCircleMidpointDistance, iNrLaps, iTVprob, dGamma,  1,
						dSimSemiMajorAxis, iEllipse1Orientation, iEllipse2Orientation,
						dSimVarianceComponent1, dSimVarianceComponent2, dCovariance, dSmoothDGP, iSimulationType);
													  
		// estimation
		vResult = fMaximumLikelihood(iNrMixtureComponentsForEstimation, 1, s_iPrint,
									 dInverseDegreesOfFreedomForEstimation, dSmoothEstimation);

		// store results
		mResult |= vResult;
	}

	
	// save result files
	sName = sprint(sResultsPath, "/Sim_", sResultsPath, "_", iSimulationType, 
		"_",iN,"_",iT,"_", dGamma,"_",
		dCircleRadius,"_",dCircleMidpointDistance,"_",
		dInverseDegreesOfFreedomForSimulation,"_",dInverseDegreesOfFreedomForEstimation,"_",
		iNrMixtureComponentsForEstimation,"_",iNrMixtureComponentsForSimulation,"_",iTVprob,
		"_", s_iMeanSpecificA_Parameters, "_", dSimSemiMajorAxis, "_",
		iEllipse1Orientation, "_", iEllipse2Orientation, "_",
		dSimVarianceComponent1, "_", dSimVarianceComponent2,  "_", dCovariance,
		"_", dSmoothEstimation, "_", dSmoothDGP);
	savemat(sprint(sName,".csv"),mResult');

	
	// Just print this number to be reminded it is there
	println("\n\n\nJust to remember it is there: regularization epsilon = ",s_dRegularizeCovariance);	
}
																																				   
